//+------------------------------------------------------------------+
//|                                                          VRT.mqh |
//|                                  Copyright 2024, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2024, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
#include<TestUtilities.mqh>
#include<Math\Alglib\specialfunctions.mqh>
#include<VectorMatrixTools.mqh>
//+------------------------------------------------------------------+
//| CVarianceRatio  class                                            |
//| Variance ratio hypthesis test for a random walk                  |
//+------------------------------------------------------------------+
class CVarianceRatio
{
 private:
   double m_pvalue;     //pvalue
   double m_statistic;  //test statistic
   double m_variance;   //variance
   double m_vr;         //variance ratio
   vector m_critvalues; //critical values 
   
 public:
   CVarianceRatio(void);
   ~CVarianceRatio(void);
 
 bool Vrt(const double &in_data[], ulong lags , ENUM_TREND trend = TREND_CONST_ONLY, bool debiased=true, bool robust=true, bool overlap = true);
 bool Vrt(const vector &in_vect, ulong lags , ENUM_TREND trend = TREND_CONST_ONLY, bool debiased=true, bool robust=true, bool overlap = true);
 
 double Pvalue(void) { return m_pvalue;}
 double Statistic(void){ return m_statistic;}
 double Variance(void){ return m_variance;}
 double VRatio(void){ return m_vr;}
 vector CritValues(void){ return m_critvalues;}
};
//+------------------------------------------------------------------+
//| constructor                                                      |
//+------------------------------------------------------------------+
CVarianceRatio::CVarianceRatio(void)
{
 m_pvalue = m_statistic = m_vr = m_variance = EMPTY_VALUE;
 
 m_critvalues.Init(6);
 
 m_critvalues[0]=0.01;
 m_critvalues[1]=0.05;
 m_critvalues[2]=0.1;
 m_critvalues[3]=0.9;
 m_critvalues[4]=0.95;
 m_critvalues[5]=0.99;
 
 for(ulong i=0; i<m_critvalues.Size(); i++)
     m_critvalues[i] = CNormalDistr::InvNormalCDF(m_critvalues[i]);

}
//+------------------------------------------------------------------+
//| Desctructor                                                      |
//+------------------------------------------------------------------+
CVarianceRatio::~CVarianceRatio(void)
{
}
//+------------------------------------------------------------------+
//| main method for computing Variance ratio test                    |
//+------------------------------------------------------------------+
bool CVarianceRatio::Vrt(const vector &in_vect,ulong lags,ENUM_TREND trend=1,bool debiased=true,bool robust=true,bool overlap=true)
{
 ulong nobs = in_vect.Size();
 
 vector y = vector::Zeros(2),delta_y;
 
 double mu;
 
 ulong nq = nobs - 1;
 
 if(in_vect.Size()<1)
   {
    Print(__FUNCTION__, "Invalid input, no data supplied");
    return false;
   }
   
 if(lags<2 || lags>=in_vect.Size())
  {
   Print(__FUNCTION__," Invalid input for lags");
   return false;
  } 
    
  if(!overlap)
   {
     if(nq % lags != 0)
        {
         ulong extra = nq%lags;
         if(!y.Init(5,slice,in_vect,0,in_vect.Size()-extra-1))
           {
            Print(__FUNCTION__," ",__LINE__);
            return false;
           }
         Print("Warning:Invalid length for input data, size is not exact multiple of lags");
        }
   } 
  else
    y.Copy(in_vect);
    
  nobs = y.Size();
  
  if(trend == TREND_NONE)
     mu = 0;
  else
     mu = (y[y.Size()-1] - y[0])/double(nobs - 1);      
  
  delta_y = difference(y);
  
  nq = delta_y.Size();
  
  vector mudiff = delta_y - mu;
  
  vector mudiff_sq = MathPow(mudiff,2.0);
  
  double sigma2_1 = mudiff_sq.Sum()/double(nq);
  
  double sigma2_q;
  
  vector delta_y_q;
  
  if(!overlap)
   {
     vector y1,y2;
     if(!y1.Init(3,slice,y,lags,y.Size()-1,lags) ||
        !y2.Init(3,slice,y,0,y.Size()-lags-1,lags))
     { 
      Print(__FUNCTION__," ",__LINE__);
      return false;
     }
     
     delta_y_q = y1-y2;
     
     vector delta_d = delta_y_q - double(lags) * mu;
     
     vector delta_d_sqr = MathPow(delta_d,2.0);
     
     sigma2_q = delta_d_sqr.Sum()/double(nq);
   }
  else
   {
     vector y1,y2;
     if(!y1.Init(3,slice,y,lags,y.Size()-1) ||
        !y2.Init(3,slice,y,0,y.Size()-lags-1))
     { 
      Print(__FUNCTION__," ",__LINE__);
      return false;
     } 
     
     delta_y_q = y1-y2;
     
     vector delta_d = delta_y_q - double(lags) * mu;
     
     vector delta_d_sqr = MathPow(delta_d,2.0);
     
     sigma2_q = delta_d_sqr.Sum()/double(nq*lags);
   }
  
  
  if(debiased && overlap)
   {
    sigma2_1 *= double(nq)/double(nq-1);
    double mm = (1.0-(double(lags)/double(nq)));
    double m = double(lags*(nq - lags+1));// * (1.0-double(lags/nq));
    sigma2_q *= double(nq*lags)/(m*mm);
   }  
  
  if(!overlap)
   m_variance = 2.0 * (lags-1);
  else
    if(!robust)
      m_variance = double((2 * (2 * lags - 1) * (lags - 1)) / (3 * lags));
    else
      {
        vector z2, o, p;
        z2=MathPow((delta_y-mu),2.0);
        double scale = pow(z2.Sum(),2.0);
        double theta = 0;
        double delta;
        for(ulong k = 1; k<lags; k++)
          {
           if(!o.Init(3,slice,z2,k,z2.Size()-1) ||
              !p.Init(3,slice,z2,0,z2.Size()-k-1))
            { 
             Print(__FUNCTION__," ",__LINE__);
             return false;
            }  
           o*=double(nq);
           p/=scale;
           delta = o.Dot(p);
           theta+=4.0*pow((1.0-double(k)/double(lags)),2.0)*delta; 
          }
        m_variance = theta;
      } 
    m_vr = sigma2_q/sigma2_1;
    
    m_statistic = sqrt(nq) * (m_vr - 1)/sqrt(m_variance);
    
    double abs_stat = MathAbs(m_statistic);
    
    m_pvalue = 2 - 2*CNormalDistr::NormalCDF(abs_stat);
    
 return true;          
}
//+------------------------------------------------------------------+
//|variance ratio test with data in array                            |
//+------------------------------------------------------------------+
bool CVarianceRatio::Vrt(const double &in_data[],ulong lags,ENUM_TREND trend=1,bool debiased=true,bool robust=true,bool overlap=true)
{
 vector series;
 
 if(series.Assign(in_data))
    return Vrt(series,lags,trend,debiased,robust,overlap);
 else
    return false;   
}